<?php

namespace Drupal\openstack\Service\Rest;

use Drupal\openstack\Service\OpenStackBatchOperations;
use Drupal\openstack\Service\OpenStackServiceInterface;
use Drupal\aws_cloud\Entity\Ec2\AwsCloudEntityInterface;
use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceBase;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Queue\QueueFactory;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Interacts with OpenStack using Native API.
 */
class OpenStackService extends CloudServiceBase implements OpenStackServiceInterface {

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Cloud context string.
   *
   * @var string
   */
  protected $cloudContext;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Guzzle client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The lock interface.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Credentials array.
   *
   * @var array
   */
  protected $credentials;

  /**
   * Constructs a new OpenStackService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   An entity type manager instance.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Http\ClientFactory $client_factory
   *   The http client.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The lock interface.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   The queue factory.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              ClientFactory $client_factory,
                              LockBackendInterface $lock,
                              QueueFactory $queue_factory) {

    // The parent constructor takes care of $this->messenger object.
    parent::__construct();

    // Setup the entity type manager for querying entities.
    $this->entityTypeManager = $entity_type_manager;

    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->httpClient = $client_factory->fromOptions();
    $this->lock = $lock;
    $this->queueFactory = $queue_factory;
  }

  /**
   * Helper method to call API.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  public function getClient(array $credentials = []) {
    $http_method = $credentials['http_method'] ?? '';
    $endpoint = $credentials['endpoint'] ?? '';
    $data = $credentials['params'] ?? '';

    if (!empty($data)) {
      $params['body'] = json_encode($data);
    }

    $this->loadCredentials();
    $auth_token = $this->generateAuthToken();
    $params['headers']['X-Auth-Token'] = $auth_token;

    $params['headers']['Content-Type'] = !empty($credentials['Content-Type'])
      ? $credentials['Content-Type'] : 'application/json';

    return $this->request($http_method, $endpoint, $params);
  }

  /**
   * {@inheritdoc}
   */
  public function setCloudContext($cloud_context) {
    $this->cloudContext = $cloud_context;
    $this->cloudConfigPluginManager->setCloudContext($cloud_context);
  }

  /**
   * Load credentials.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   The credentials array.
   */
  private function loadCredentials(array $credentials = []) {
    if (!empty($credentials)) {
      $this->credentials = $credentials;
    }
    elseif (empty($this->credentials)) {
      $this->credentials = $this->cloudConfigPluginManager->loadCredentials();
    }

    return $this->credentials;
  }

  /**
   * Setup any default parameters for the Guzzle request.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array of parameters.
   */
  public function generateAuthToken(array $credentials = []) {
    $this->loadCredentials($credentials);

    $body = [
      'auth' => [
        'identity' => [
          'methods' => [0 => 'password'],
          'password' => [
            'user' => [
              'name' => $this->credentials['username'],
              'domain' => [
                'id' => $this->credentials['domain_id'],
              ],
              'password' => $this->credentials['password'],
            ],
          ],
        ],
        'scope' => [
          'project' => [
            'id' => $this->credentials['project_id'],
          ],
        ],
      ],
    ];

    $response = $this->request('POST', '/identity/v3/auth/tokens', ['json' => $body]);

    $result = $response['auth_token'] ?? $response;
    return $result;
  }

  /**
   * Get Instance List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance list API response.
   */
  public function describeInstances(array $params = []) {
    $credentials = [
      'http_method' => 'get',
      'endpoint' => '/compute/v2.1/servers/detail',
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $results = $private_ips = $public_ips = [];

    foreach ($response['body']['servers'] ?: [] as $key => $instance) {
      $results['Reservations'][$key]['Groups'] = [];

      if (!empty($instance['security_groups'])) {
        foreach ($instance['security_groups'] ?: [] as $key_group => $group) {
          $results['Reservations'][$key]['Groups'][$key_group]['GroupName'] = $group['name'] ?? NULL;
        }
      }

      // Get Volume Response.
      if (!empty($instance['os-extended-volumes:volumes_attached']['0']['id'])) {
        $volume_data = $this->describeVolumeById([], $instance['os-extended-volumes:volumes_attached']['0']['id']);
        $instance['image'] = empty($instance['image'])
          ? ['id' => $volume_data['volume_image_metadata']['image_id']] : [];
      }

      // Get flavor Response.
      $flavor_id = $instance['flavor']['id'] ?? NULL;
      $flavor_data = $this->describeFlavorById([], $flavor_id);

      $results['Reservations'][$key]['OwnerId'] = $instance['tenant_id'];

      $results['Reservations'][$key]['Instances'][0] = [
        'AmiLaunchIndex' => NULL,
        'ImageId' => $instance['image']['id'] ?? NULL,
        'InstanceId' => $instance['id'] ?? NULL,
        'InstanceName' => $instance['name'] ?? NULL,
        'InstanceType' => $flavor_data['name'] ?? NULL,
        'KeyName' => $instance['key_name'] ?? NULL,
        'LaunchTime' => str_replace("T", " ", $instance['OS-SRV-USG:launched_at']) ?? NULL,
        'Placement' => ['AvailabilityZone' => $instance['OS-EXT-AZ:availability_zone']] ?? NULL,
        'PrivateDnsName' => $instance['name'] ?? NULL,
        'PrivateIpAddress' => $instance['addresses']['private'][1]['addr'] ?? NULL,
        'PublicIpAddress' => $instance['addresses']['shared'][1]['addr'] ?? NULL,
        'PublicDnsName' => $instance['dns_name'] ?? NULL,
        'State' => ['Name' => $instance['OS-EXT-STS:power_state'] === 1 ? 'running' : 'stopped'],
        'RootDeviceName' => NULL,
        'RootDeviceType' => NULL,
        'SecurityGroups' => $results['Reservations'][$key]['Groups'] ?? [],
      ];
    }

    return $results;
  }

  /**
   * Get Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image list API response.
   */
  public function describeImages(array $params = []) {

    $credentials = [
      'http_method' => 'get',
      'endpoint' => '/image/v2/images',
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    $results = [];

    foreach ($response['body']['images'] as $key => $value) {
      $results['Images'][$key]['Architecture'] = NULL;
      $results['Images'][$key]['CreationDate'] = $value['created_at'];
      $results['Images'][$key]['ImageId'] = $value['id'];
      $results['Images'][$key]['ImageLocation'] = NULL;
      $results['Images'][$key]['ImageType'] = $value['disk_format'];
      $results['Images'][$key]['Public'] = $value['visibility'] == 'public' ? 1 : 0;
      $results['Images'][$key]['OwnerId'] = $value['owner'];
      $results['Images'][$key]['State'] = $value['status'] == 'active' ? 'available' : 'pending';
      $results['Images'][$key]['Name'] = $value['name'];
      $results['Images'][$key]['RootDeviceType'] = $value['container_format'];
    }

    return $results;
  }

  /**
   * Create Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image create API response.
   */
  public function createImage(array $params = []) {

    $body = [
      'container_format' => 'bare',
      'disk_format' => 'qcow2',
      'name' => $params['Name'],
      'visibility' => $params['Visibility'] == 0 ? 'private' : 'public',
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => "/image/v2/images",
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $results = [];

    $results['ImageId'] = $response['body']['id'];
    $results['name'] = $response['body']['name'];
    $results['disk_format'] = $response['body']['disk_format'];
    $results['container_format'] = $response['body']['container_format'];
    $results['visibility'] = $response['body']['visibility'];
    $results['created_at'] = $response['body']['created_at'];
    $results['updated_at'] = $response['body']['updated_at'];
    $results['tags'] = $response['body']['tags'];
    $results['status'] = $response['body']['status'];

    return $results;
  }

  /**
   * Update Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image update API response.
   */
  public function updateImage(array $params = []) {
    $body = [
      [
        'op' => 'replace',
        'path' => '/name',
        'value' => $params['Name'],
      ],
      [
        'op' => 'replace',
        'path' => '/visibility',
        'value' => $params['Visibility'],
      ],
    ];

    $credentials = [
      'http_method' => 'patch',
      'endpoint' => "/image/v2/images/{$params['ImageId']}",
      'params' => $body,
    ];

    $credentials['Content-Type'] = 'application/openstack-images-v2.1-json-patch';

    $response = $this->getClient($credentials);

    $results = [];

    $results['ImageId'] = $response['body']['id'];
    $results['name'] = $response['body']['name'];
    $results['disk_format'] = $response['body']['disk_format'];
    $results['container_format'] = $response['body']['container_format'];
    $results['visibility'] = $response['body']['visibility'];
    $results['created_at'] = $response['body']['created_at'];
    $results['updated_at'] = $response['body']['updated_at'];
    $results['tags'] = $response['body']['tags'];
    $results['status'] = $response['body']['status'];

    return $results;
  }

  /**
   * Delete Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image delete API response.
   */
  public function deregisterImage(array $params = []) {
    $credentials = [
      'http_method' => 'DELETE',
      'endpoint' => "/image/v2/images/{$params['ImageId']}",
    ];

    $response = $this->getClient($credentials);

    if ($response['status_code'] === 204) {
      return TRUE;
    }
  }

  /**
   * Get Flavor Detail of specific flavor Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $flavor_id
   *   The Flavor Id.
   *
   * @return array
   *   The Flavor list API response.
   */
  public function describeFlavorById(array $params = [], $flavor_id = '') {

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/compute/v2.1/flavors/{$flavor_id}",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['flavor'];
  }

  /**
   * Get Volume Detail of specific volume Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $volume_id
   *   The Volume Id.
   *
   * @return array
   *   The Volume list API response.
   */
  public function describeVolumeById(array $params = [], $volume_id = '') {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => "/volume/v3/{$project_id}/volumes/{$volume_id}",
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['volume'];
  }

  /**
   * {@inheritdoc}
   */
  public function describeAvailabilityZones(array $params = []) {
    $credentials = [
      'http_method' => 'get',
      'endpoint' => '/compute/v2.1/os-availability-zone',
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['availabilityZoneInfo'];
  }

  /**
   * Call the API for updated entities and store them as Instance entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateInstanceEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '') {
    $updated = FALSE;
    // Call the api and get all instances.
    $result = $this->describeInstances($params);
    if ($result !== NULL) {
      $all_instances = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the instances by setting up
      // the array with the instance_id.
      foreach ($all_instances ?: [] as $instance) {
        $stale[$instance->getInstanceId()] = $instance;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Instance Update');
      // Loop through the reservations and store each one as an Instance entity.
      foreach ($result['Reservations'] ?: [] as $reservation) {

        foreach ($reservation['Instances'] ?: [] as $instance) {
          // Keep track of instances that do not exist anymore
          // delete them after saving the rest of the instances.
          if (isset($stale[$instance['InstanceId']])) {
            unset($stale[$instance['InstanceId']]);
          }
          // Store the Reservation OwnerId in instance so batch
          // callback has access.
          $instance['reservation_ownerid'] = $reservation['OwnerId'] ?? NULL;
          $instance['reservation_id'] = $reservation['ReservationId'] ?? NULL;

          $batch_builder->addOperation([
            OpenStackBatchOperations::class,
            'updateInstance',
          ], [$cloud_context, $instance]);

        }
      }
      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Update the Instances.
   *
   * @param array $params
   *   Parameters array to send to API.
   * @param bool $clear
   *   TRUE to clear stale instances.
   *
   * @return bool
   *   Indicates success or failure.
   */
  public function updateInstances(array $params = [], $clear = TRUE) {
    $entity_type = 'openstack_instance';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateInstanceEntities($params, $clear, $entity_type, $this->cloudContext);
    $this->lock->release($lock_name);

    return $updated;
  }

  /**
   * Update the OpenStackImages.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateImages(array $params = [], $clear = TRUE) {
    $entity_type = 'openstack_image';
    $lock_name = $this->getLockKey($entity_type);
    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateImageEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as Image entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateImageEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '') {
    $updated = FALSE;

    // Load all entities by $this->cloudContext.
    $image_entities = $this->entityTypeManager->getStorage($entity_type)->loadByProperties(
      ['cloud_context' => $this->cloudContext]
    );

    $result = $this->describeImages($params);

    if (isset($result)) {

      $stale = [];
      // Make it easier to lookup the images by setting up
      // the array with the image_id.
      foreach ($image_entities ?: [] as $image) {
        $stale[$image->getImageId()] = $image;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Image Update');

      foreach ($result['Images'] ?: [] as $image) {
        // Keep track of images that do not exist anymore
        // delete them after saving the rest of the images.
        if (isset($stale[$image['ImageId']])) {
          unset($stale[$image['ImageId']]);
        }

        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateImage',
        ], [$cloud_context, $image]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailabilityZones($with_id = FALSE) {
    $zones = [];
    $result = $this->describeAvailabilityZones();
    if (!empty($result)) {
      foreach ($result ?: [] as $zone_info) {
        if ($with_id) {
          $name = sprintf('%s (%s)', $zone_info['zoneName'], $zone_info['zoneId']);
        }
        else {
          $name = $zone_info['zoneName'];
        }
        $zones[$zone_info['zoneName']] = $name;
      }
    }
    return $zones;
  }

  /**
   * Helper method to do a Guzzle HTTP Request.
   *
   * @param string $http_method
   *   The http method in lower case.
   * @param string $endpoint
   *   The endpoint to access.
   * @param array $params
   *   Parameters to pass to the request.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  private function request($http_method, $endpoint, array $params = []) {
    $output = [];

    try {
      $response = $this->httpClient->$http_method(
        rtrim($this->credentials['authUrl'], '/') . $endpoint,
        $params
      );

      if (!empty($response)) {
        $status_code = $response->getStatusCode();
        $headers = $response->getHeaders();
        $output['auth_token'] = $headers['X-Subject-Token'][0] ?? '';
        $output['status_code'] = $status_code;
        $body = $response->getBody()->getContents();
      }
    }
    catch (GuzzleException $error) {
      $response = $error->getResponse();
      $content = '';
      if (!empty($response)) {
        $content = $response->getBody()->getContents();
        $status_code = $response->getStatusCode();
        $output['status_code'] = $status_code;
      }

      $this->messenger->addError($this->t('OpenStack API error. Error details are as follows: @response', [
        '@response' => $error->getMessage(),
      ]));

      $message = new FormattableMarkup(
        'OpenStack API error. Error details are as follows:<pre>@response</pre>', [
          '@response' => !empty($content)
          ? $content
          : $error->getMessage(),
        ]
      );
      $this->logError('Remote API Connection', $error, $message, FALSE);
    }
    catch (\Exception $error) {
      $this->messenger->addError($this->t('An unknown error occurred while
      trying to connect to the remote API. This is not a Guzzle error, not an
      error in the remote API, rather a generic local error occurred.
      The reported error was @error', [
        '@error' => $error->getMessage(),
      ]));

      $this->logError('Remote API Connection', $error, $this->t('An unknown error
      occurred while trying to connect to the remote API. This is not a Guzzle
      error, nor an error in the remote API, rather a generic local error
      occurred. The reported error was @error',
        ['@error' => $error->getMessage()]
      ), FALSE);
    }
    $body = empty($body) ? [] : json_decode($body, TRUE);
    $output['body'] = $body;
    return $output;
  }

  /**
   * Generate a lock key based on entity name.
   *
   * @param string $name
   *   The entity name.
   *
   * @return string
   *   The lock key.
   */
  protected function getLockKey($name) {
    return $this->cloudContext . '_' . $name;
  }

  /**
   * Initialize a new batch builder.
   *
   * @param string $batch_name
   *   The batch name.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The initialized batch object.
   */
  protected function initBatch($batch_name) {
    return (new BatchBuilder())
      ->setTitle($batch_name);
  }

  /**
   * Run the batch job to process entities.
   *
   * @param \Drupal\Core\Batch\BatchBuilder $batch_builder
   *   The batch builder object.
   */
  protected function runBatch(BatchBuilder $batch_builder) {
    try {
      // Log the start time.
      $start = $this->getTimestamp();
      $batch_array = $batch_builder->toArray();
      batch_set($batch_array);
      // Reset the progressive so batch works with out a web head.
      $batch = &batch_get();
      $batch['progressive'] = FALSE;
      // Check if using Drush to process the queue.
      // using drush_backend_batch_process() will fix the "Route not found"
      // error.
      if (PHP_SAPI === 'cli' && function_exists('drush_backend_batch_process')) {
        drush_backend_batch_process();
      }
      else {
        batch_process();
      }
      // Log the end time.
      $end = $this->getTimestamp();
      $this->logger('openstack_rest')->info($this->t('@updater - @cloud_context: Batch operation took @time seconds.', [
        '@cloud_context' => $this->cloudContext,
        '@updater' => $batch_array['title'],
        '@time' => $end - $start,
      ]));
    }
    catch (Exception $e) {
      $this->handleException($e);
    }
    finally {
      // Reset the batch otherwise this operation hangs when using Drush.
      // https://www.drupal.org/project/drupal/issues/3166042
      $batch = [];
    }
  }

  /**
   * Log errors to watchdog and throw an exception.
   *
   * @param string $type
   *   The error type.
   * @param \Exception $exception
   *   The exception object.
   * @param string $message
   *   The message to log.
   * @param bool $throw
   *   TRUE to throw exception.
   *
   * @throws \Drupal\vmware\Service\VmwareServiceException
   */
  private function logError($type, \Exception $exception, $message, $throw = TRUE) {
    watchdog_exception($type, $exception, $message);
    if ($throw === TRUE) {
      throw new OpenStackServiceException($message);
    }
  }

  /**
   * Helper static method to clear cache.
   */
  public static function clearCacheValue() {
    \Drupal::cache('menu')->invalidateAll();
    \Drupal::service('cache.render')->deleteAll();
    \Drupal::service('router.builder')->rebuild();
    \Drupal::service('plugin.cache_clearer')->clearCachedDefinitions();
  }

  /**
   * Helper method to get the current timestamp.
   *
   * @return int
   *   The current timestamp.
   */
  protected function getTimestamp() {
    return time();
  }

  /**
   * Get Tag name.
   *
   * @param array $openstack_obj
   *   The openstack object.
   * @param string $default_value
   *   The default value.
   *
   * @return string
   *   Return name of corresponding resource.
   */
  public function getTagName(array $openstack_obj, $default_value) {
    $name = $default_value;
    if (empty($openstack_obj['name'])) {
      return $name;
    }

    if (!empty($openstack_obj['name'])) {
      return $openstack_obj['name'];
    }

    foreach ($openstack_obj['Tags'] ?: [] as $tag) {
      if ($tag['Key'] === 'Name' && !empty($tag['Value'])) {
        $name = $tag['Value'];
        break;
      }
    }
    return $name;
  }

  /**
   * Method to clear all entities out of the system.
   */
  public function clearAllEntities() {
    $timestamp = $this->getTimestamp();
    $this->clearEntities('openstack_instance', $timestamp);
    $this->clearEntities('openstack_image', $timestamp);
  }

  /**
   * {@inheritdoc}
   */
  public function createResourceQueueItems(CloudConfigInterface $cloud_config) {
    $update_resources_queue = $this->queueFactory->get('openstack_update_resources_queue');
    $method_names = [
      'updateInstances',
      'updateImages',
    ];
    foreach ($method_names as $method_name) {
      $update_resources_queue->createItem([
        'cloud_context' => $this->cloudContext,
        'openstack_method_name' => $method_name,
      ]);
    }
  }

  /**
   * Get UID tag value.
   *
   * @param array $tags_array
   *   The tags array of API.
   * @param string $entity_type_id
   *   The entity type id.
   *
   * @return int
   *   Return user id.
   */
  public function getUidTagValue(array $tags_array, $entity_type_id) {
    $key = $entity_type_id . '_' . AwsCloudEntityInterface::TAG_CREATED_BY_UID;
    $uid = 0;
    if (!empty($tags_array['Tags'])) {
      foreach ($tags_array['Tags'] ?: [] as $tag) {
        if ($tag['Key'] === $key) {
          $uid = $tag['Value'];
          break;
        }
      }
    }
    return $uid;
  }

}
